/*
 * Ctunnel - Cyptographic tunnel.
 * Copyright (C) 2008-2011 Jess Mahan <ctunnel@nardcore.org>
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <pthread.h>

#include "config.h"
#include "ctunnel.h"
#include "comp.h"
#include "crypto.h"
#include "net.h"
#include "opt.h"
#include "thread.h"

void crypto_shutdown(int sig)
{
    ctunnel_log(stdout, LOG_INFO, "Shutting Down\n");
#ifdef HAVE_OPENSSL
    EVP_cleanup();
#endif
    exit(0);
}

int main(int argc, char *argv[])
{
    int srv_sockfd = 0;
    int ret = 0;
    int i = 0, x = 0;
#ifndef _WIN32
    struct sockaddr_in pin;
    socklen_t addrsize = 0;
    pid_t pid;
#endif
//    void *data = NULL;
    struct options opt;
    struct Network *net_cli = NULL, *net_srv = NULL;
    //struct Ctunnel ctunnel[MAX_THREADS];
    struct Ctunnel **ctunnel;
    pthread_t tid[MAX_THREADS];
    extern int threads[MAX_THREADS];
    struct timeval tv;

    for (i = 0; i < MAX_THREADS; i++)
	threads[i] = 0;
#ifndef _WIN32
    for (i = 0; i < MAX_THREADS; i++)
	tid[i] = (pthread_t) 0;
#endif

    pthread_mutex_init(&mutex, NULL);

    ctunnel = malloc(sizeof(struct Ctunnel **) * MAX_THREADS);

    i = 0;

    fprintf(stdout, "%s %d.%d %s\n", SOFTWARE_NAME, VERSION_MAJOR,
	    VERSION_MINOR, COPYRIGHT);

#ifndef _WIN32
    openlog("ctunnel", LOG_NDELAY | LOG_PID, LOG_DAEMON);
    ctunnel_log(stdout, LOG_INFO, "ctunnel %d.%d starting", VERSION_MAJOR,
		VERSION_MINOR);
#endif

    memset(opt.host, 0x00, sizeof(opt.host));
    opt = get_options(argc, argv);

#ifndef _WIN32
    signal(SIGCHLD, SIG_IGN);
    signal(SIGINT, crypto_shutdown);
    signal(SIGTERM, crypto_shutdown);
#endif

    if (opt.proto == 0) {
	srv_sockfd = tcp_listen(opt.port_listen);
	net_cli = calloc(1, sizeof(struct Network));
	net_srv = calloc(1, sizeof(struct Network));
    } else {
	net_srv = udp_bind(opt.port_listen);
	net_cli = udp_connect(opt.host, opt.port_forward);
    }

    if (opt.proto == 0) {
#ifdef HAVE_OPENSSL
	ctunnel_log(stdout, LOG_INFO,
		    "TCP [OpenSSL] Listening on %d using %s",
		    opt.port_listen, opt.key.ciphername);
#else
	ctunnel_log(stdout, LOG_INFO,
		    "TCP [libgcrypt] Listening on %d using %s %s",
		    opt.port_listen, opt.key.ciphername, opt.mode);
#endif
    } else {
#ifdef HAVE_OPENSSL
	ctunnel_log(stdout, LOG_INFO,
		    "UDP [OpenSSL] Listening on %d using %s",
		    opt.port_listen, opt.key.ciphername);
#else
	ctunnel_log(stdout, LOG_INFO,
		    "UDP [libgcrypt] Listening on %d using %s %s",
		    opt.port_listen, opt.key.ciphername, opt.mode);
#endif
    }

    if (opt.comp == 1)
	ctunnel_log(stdout, LOG_INFO,
		    "zlib compression enabled [level %d]", opt.comp_level);

    if (srv_sockfd < 0) {
	ctunnel_log(stderr, LOG_CRIT, "listen() %s\n", strerror(errno));
	exit(1);
    }
#ifndef _WIN32
    if (opt.daemon != 1) {
	ctunnel_log(stdout, LOG_INFO, "Going into background\n");
	pid = fork();
	if (pid == -1) {
	    ctunnel_log(stdout, LOG_CRIT, "fork() %s\n", strerror(errno));
	    exit(1);
	}
	if (pid > 0) {
	    exit(0);
	}
    }
#endif

    tv.tv_sec = 5;
    tv.tv_usec = 0;

    //for (i = 0;; i++) {
    while (1) {
	if (opt.proto == 0) {
	    setsockopt(srv_sockfd, SOL_SOCKET, SO_RCVTIMEO,
		       (struct timeval *) &tv, sizeof(struct timeval));
	    setsockopt(srv_sockfd, SOL_SOCKET, SO_SNDTIMEO,
		       (struct timeval *) &tv, sizeof(struct timeval));
	    setsockopt(net_cli->sockfd, SOL_SOCKET, SO_RCVTIMEO,
		       (struct timeval *) &tv, sizeof(struct timeval));
	    setsockopt(net_cli->sockfd, SOL_SOCKET, SO_SNDTIMEO,
		       (struct timeval *) &tv, sizeof(struct timeval));
	    setsockopt(net_srv->sockfd, SOL_SOCKET, SO_RCVTIMEO,
		       (struct timeval *) &tv, sizeof(struct timeval));
	    setsockopt(net_srv->sockfd, SOL_SOCKET, SO_SNDTIMEO,
		       (struct timeval *) &tv, sizeof(struct timeval));
#ifdef _WIN32
	    net_srv->sockfd = accept(srv_sockfd, NULL, NULL);
#else
	    net_srv->sockfd = accept(srv_sockfd, (struct sockaddr *) &pin,
				     &addrsize);
#endif
	    if (net_srv->sockfd > 0) {
		net_cli->sockfd = tcp_connect(opt.host, opt.port_forward);

		if (net_cli->sockfd > 0) {

		    pthread_mutex_lock(&mutex);
		    for (i = 0; i != MAX_THREADS; i++) {
			if (threads[i] == 0)
			    break;
			if (i == MAX_THREADS)
			    ctunnel_log(stderr, LOG_CRIT,
					"Max Threads %d " \
				        "reached!", i);
		    }
		    //fprintf(stdout, "THREAD: Start %d %d/%d\n",
		    //	    i, net_cli->sockfd, net_srv->sockfd);
		    ctunnel[i] = malloc(sizeof(struct Ctunnel));
		    ctunnel[i]->clisockfd = net_cli->sockfd;
		    ctunnel[i]->srvsockfd = net_srv->sockfd;
		    ctunnel[i]->opt = opt;
		    ctunnel[i]->srv_sockfd = srv_sockfd;
		    ctunnel[i]->id = i;
		    pthread_mutex_unlock(&mutex);

		    if (opt.proto == 0) {
			threads[i] = 1;
			ret = pthread_create(&tid[i], NULL,
					     (void *) ctunnel_mainloop,
					     (void *) ctunnel[i]);
			//pthread_detach(tid[i]);
		    } else {
			ctunnel[i]->net_cli = net_cli;
			ctunnel[i]->net_srv = net_srv;
			ctunnel_mainloop((void *) ctunnel[i]);
		    }
		} else {
		    net_close(net_srv);
		}
	    }
	}
	for (x = 0; x != MAX_THREADS; x++) {
	    if (threads[x] == 2) {
		//fprintf(stdout, "THREAD: Join %d\n", x);
		pthread_join(tid[x], NULL);
		pthread_mutex_lock(&mutex);
		threads[x] = 0;
		pthread_mutex_unlock(&mutex);
	    }
	}
	/*
	fprintf(stdout, "THREAD: In use: ");
	for (x = 0; x != MAX_THREADS; x++) {
	    if (threads[x] == 1)
		fprintf(stdout, "%d ", x);
	}
	fprintf(stdout, "\n");
	*/
    }
    return 0;
}
