/*
 * Ctunnel - Cyptographic tunnel.
 * Copyright (C) 2008-2011 Jess Mahan <ctunnel@nardcore.org>
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include "config.h"
#include "ctunnel.h"

int bin_size(unsigned char *s)
{
    unsigned char *p = s;

    while (*p)
	p++;

    return p - s;
}

void version(void)
{
    fprintf(stdout, "%s %d.%d %s\n", SOFTWARE_NAME, VERSION_MAJOR,
	    VERSION_MINOR, COPYRIGHT);
}

void usage(void)
{
    fprintf(stdout, "Usage is as follows:\n"
#ifdef HAVE_OPENSSL
	    "./ctunnel [-U][-n][-p][-v][-h][-b][-z level] -c|-s -l -f -H -C cipher\n"
	    "Using OpenSSL for crypto\n"
#else
	    "./ctunnel [-U][-n][-p][-v][-h][-z level] -c|-s -l -f -H -C cipher -M mode\n"
	    "Using libgcrypt for crypto\n"
#endif
	    "  -U\t(optional) Use UDP (default is TCP)\n"
	    "  -n\t(optional) Do not fork into Background\n"
	    "  -p\t(optional) Print Stored Key, IV, and Cipher then exit\n"
	    "  -v\t(optional) Print Version information then exit\n"
	    "  -h\t(optional) Print the screen you are reading then exit\n"
	    "  -z #  (optional) Enable Compression at level #, 5 default.\n"
	    "  -b #  (optional) Packet buffer size in bytes (%d) by default.\n"
	    "  -c\t(manditory) Operate in Client Mode (do not use with -s)\n"
	    "  -s\t(manditory) Operate in Server Mode (do not use with -c)\n"
	    "  -l\t(manditory) Listen for TCP connections on this port\n"
	    "  -f\t(manditory) Forward TCP connection from -l to this port\n"
	    "  -H\t(manditory) Host to forward TCP connections to\n"
	    "  -C\t(manditory) Encrypt TCP traffic with this cipher\n"
#ifndef HAVE_OPENSSL
	    "  -M\t(manditory) Cipher Mode (ecb,cfb,cbc,ofb) cfb is recommended\n\n"
#endif
	    "Example: (Tunnel Mysql)\n" "  Client Machine:\n"
#ifdef HAVE_OPENSSL
	    "    ./ctunnel -c -l 2020 -f 2021 -H remote.ip -C aes-256-cfb\n"
	    "  Remote Machine:\n"
	    "    ./ctunnel -s -l 2021 -f 3306 -H 127.0.0.1 -C aes-256-cfb\n"
#else
	    "    ./ctunnel -c -l 2020 -f 2021 -H remote.ip -C aes256 -M cfb\n"
	    "  Remote Machine:\n"
	    "    ./ctunnel -s -l 2021 -f 3306 -H 127.0.0.1 -C aes256 -M cfb\n"
#endif
	    "  Client Machine:\n"
	    "    mysql -u root -p -h 127.0.0.1 -P 2020\n", PACKET_SIZE);
}

struct options get_options(int argc, char *argv[])
{
    int i = 0, x = 0;
    struct options opt;
    char *homedir;
    char *passkey;
    struct stat st;
#ifndef HAVE_OPENSSL
    crypto_ctx ctx;
    int ret = 0;
#endif
    int fd;

    if (argc <= 1) {
	usage();
	exit(0);
    }

    opt.port_listen = -1;
    opt.port_forward = -1;
    opt.server = -1;
    opt.daemon = -1;
    opt.proto = 0;
    opt.comp = 0;
    opt.comp_level = 5;
    opt.packet_size = PACKET_SIZE;
    memset(opt.host, 0x00, sizeof(opt.host));
    memset(opt.key.key, 0x00, sizeof(opt.key.key));
    memset(opt.key.iv, 0x00, sizeof(opt.key.iv));
    memset(opt.cipher, 0x00, sizeof(opt.cipher));
#ifndef HAVE_OPENSSL
    memset(opt.mode, 0x00, sizeof(opt.mode));
#endif

#ifndef _WIN32
    homedir = getenv("HOME");
    passkey = calloc(1, strlen(homedir) + strlen(KEYFILE) + 2);
    snprintf(passkey, strlen(homedir) + strlen(KEYFILE) + 1, "%s%s",
	     homedir, KEYFILE);
#else
    homedir = getenv("APPDATA");
    passkey = calloc(1, sizeof(char *) * 25);
    snprintf(passkey, strlen(homedir) + 25, "%s\\ctunnel-passkey",
	     homedir);
#endif
    if (stat(passkey, &st) < 0) {
	fprintf(stdout, "No Passkey, will prompt\n");
    } else {
#ifndef _WIN32
	fd = open(passkey, O_RDONLY,
		  S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
#else
	fd = _open(passkey, O_RDONLY | O_BINARY);
#endif
	opt.key.version = CONFIG_VERSION;
	read(fd, &opt.key, sizeof(struct keys));
	strcpy(opt.cipher, opt.key.ciphername);
	if (opt.key.version != CONFIG_VERSION) {
	    fprintf(stderr, "Passkey in %s not compatible, please delete "
		            "and re-reun ctunnel\n", passkey);
	    exit(1);
	}
	ctunnel_log(stdout, LOG_INFO, "Found Stored Key and Cipher[%s] in %s",
		    opt.key.ciphername, passkey);
	close(fd);
    }
    for (i = 0; argv[i] != NULL; i++) {
	if (argv[i][0] == '-') {
	    switch (argv[i][1]) {
	    case 'p':
		fprintf(stdout, "Key:    %s\nIV:     %s\nCipher: %s\n",
			opt.key.key, opt.key.iv, opt.cipher);
		exit(0);
		break;
	    case 's':
		opt.server = 1;
		break;
	    case 'c':
		opt.server = 0;
		break;
	    case 'n':
		opt.daemon = 1;
		break;
	    case 'h':
		usage();
		exit(0);
		break;
	    case 'v':
		version();
		break;
	    case 'U':
		opt.proto = 1;
		break;
	    case 'z':
		opt.comp = 1;
		break;
	    }
	}
    }
    for (i = 0; argv[i] != NULL; i++) {
	if (argv[i][0] == '-' && argv[i + 1]) {
	    switch (argv[i][1]) {
	    case 'l':
		if (argv[i + 1][0] != '-')
		    opt.port_listen = atoi(argv[i + 1]);
		break;
	    case 'f':
		if (argv[i + 1][0] != '-')
		    opt.port_forward = atoi(argv[i + 1]);
		break;
	    case 'C':
		if (argv[i + 1][0] != '-')
		    strcpy(opt.cipher, argv[i + 1]);
		break;
#ifndef HAVE_OPENSSL
	    case 'M':
		if (argv[i + 1][0] != '-')
		    strcpy(opt.mode, argv[i + 1]);
		break;
#else
	    case 'M':
		fprintf(stderr, "Cipher and Mode are specified with -C "
			"only for OpenSSL\n");
		exit(1);
		break;
#endif
	    case 'H':
		if (argv[i + 1][0] != '-')
		    strcpy(opt.host, argv[i + 1]);
		break;
	    case 'z':
		if (argv[i + 1][0] != '-')
		    opt.comp_level = atoi(argv[i + 1]);
		break;
	    case 'b':
		if (argv[i + 1][0] != '-')
		    opt.packet_size = atoi(argv[i + 1]);
		break;
	    }
	}
    }

    if ((bin_size(opt.key.key)) < 16) {
	fprintf(stdout, "Enter Key [16 Characters]: ");
	for (x = 0; x < 16; x++) {
	    opt.key.key[x] = fgetc(stdin);
	}
	fgetc(stdin);
    }
    if ((bin_size(opt.key.iv)) < 16) {
	fprintf(stdout, "Enter IV  [16 Characters]: ");
	for (x = 0; x < 16; x++) {
	    opt.key.iv[x] = fgetc(stdin);
	}
	fgetc(stdin);
    }

    /* Check ciphers/modes */
#ifndef HAVE_OPENSSL
    if ((strlen(opt.mode)) > 0) {
	if (!strcmp(opt.mode, "ecb"))
	    opt.key.mode = GCRY_CIPHER_MODE_ECB;
	else if (!strcmp(opt.mode, "cfb"))
	    opt.key.mode = GCRY_CIPHER_MODE_CFB;
	else if (!strcmp(opt.mode, "cbc"))
	    opt.key.mode = GCRY_CIPHER_MODE_CBC;
	else if (!strcmp(opt.mode, "ofb"))
	    opt.key.mode = GCRY_CIPHER_MODE_OFB;
	else
	    opt.key.mode = 0;
    } else {
	fprintf(stdout, "ERROR: Please specify Cipher Mode with -M\n");
	exit(0);
    }

#endif

    if ((strlen(opt.cipher)) > 0) {
	strcpy(opt.key.ciphername, opt.cipher);
    }
#ifdef HAVE_OPENSSL
	OpenSSL_add_all_ciphers();
	opt.key.cipher = EVP_get_cipherbyname(opt.cipher);
	if (!opt.key.cipher) {
	    ctunnel_log(stderr, LOG_CRIT, "[OpenSSL] Cipher \"%s\" not found!", opt.cipher);
	    exit(1);
	}
	EVP_cleanup();
#else
	if ((gcry_cipher_map_name(opt.cipher)) == 0) {
	    ctunnel_log(stderr, LOG_CRIT, "[libgcrypt] Cipher \"%s\" not found!", opt.cipher);
	    exit(1);
	} 
	opt.key.cipher = gcry_cipher_map_name(opt.cipher);
	ret = gcry_cipher_open(&ctx, opt.key.cipher, opt.key.mode, 0);
	if (!ctx) {
	    ctunnel_log(stderr, LOG_CRIT, "gcry_open_cipher: %s", gpg_strerror(ret));
	    gcry_cipher_close(ctx);
	    exit(1);
	}
	gcry_cipher_close(ctx);
#endif
//    }

#ifdef _WIN32
    fd = open(passkey, O_RDWR | O_BINARY | O_CREAT);
#else
    fd = open(passkey, O_RDWR | O_TRUNC | O_CREAT,
	      S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
#endif
    opt.key.version = CONFIG_VERSION;
    write(fd, &opt.key, sizeof(struct keys));
    close(fd);

    if ((bin_size((unsigned char *) opt.key.ciphername)) <= 0) {
	fprintf(stdout, "ERROR: Please specify Stream Cipher with -C\n");
	exit(0);
    }

    if (opt.server < 0) {
	fprintf(stdout,
		"ERROR: Please specify Client of Server mode with -c or -s\n");
	exit(0);
    }
    if ((bin_size((unsigned char *) opt.host)) <= 0) {
	fprintf(stdout, "ERROR: Please specify Host with -H\n");
	exit(0);
    }
    if (opt.port_listen < 0) {
	fprintf(stdout, "ERROR: Please specify Listen Port with -l\n");
	exit(0);
    }
    if (opt.port_forward < 0) {
	fprintf(stdout, "ERROR: Please specify Forawrd Port with -f\n");
	exit(0);
    }
    if (opt.comp_level > 9 || opt.comp_level < 0) {
	fprintf(stdout, "ERROR: Compression level must be between 0 and 9\n");
	exit(0);
    }
    if (opt.packet_size < 1) {
	fprintf(stdout, "ERROR: Packet size must be greater than 0\n");
	exit(0);
    }
    free(passkey);

    return opt;
}
