/*
 * Ctunnel - Cyptographic tunnel.
 * Copyright (C) 2008-2011 Jess Mahan <ctunnel@nardcore.org>
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include <pthread.h>

#include "config.h"
#include "ctunnel.h"
#include "comp.h"
#include "crypto.h"
#include "net.h"

void *ctunnel_mainloop(void *arg)
{
    struct Ctunnel *ctunnel;
    struct Network *net_srv;
    struct Network *net_cli;
    struct Compress comp;
    struct Packet pkt_in = { NULL, 0 }, pkt_out = { NULL, 0 };
    unsigned char *data;
    fd_set rfd;
    int ret = 0;
    extern int threads[MAX_THREADS];
    struct timeval tv;
#ifdef HAVE_OPENSSL
    crypto_ctx *ectx;
    crypto_ctx *dctx;
    struct Packet (*do_encrypt) (crypto_ctx * ctx, unsigned char *intext,
				 int size, unsigned char *key,
				 unsigned char *iv);
    do_encrypt = openssl_do_encrypt;
    struct Packet (*do_decrypt) (crypto_ctx * ctx, unsigned char *intext,
				 int size, unsigned char *key,
				 unsigned char *iv);
    do_decrypt = openssl_do_decrypt;
    crypto_ctx *(*crypto_init) (struct options opt, int dir);
    crypto_init = openssl_crypto_init;
    void (*crypto_deinit) (crypto_ctx *ctx);
    crypto_deinit = openssl_crypto_deinit;
#else
    crypto_ctx ectx = NULL;
    crypto_ctx dctx = NULL;
    struct Packet (*do_encrypt) (crypto_ctx ctx, unsigned char *intext,
				 int size, unsigned char *key,
				 unsigned char *iv);
    do_encrypt = gcrypt_do_encrypt;
    struct Packet (*do_decrypt) (crypto_ctx ctx, unsigned char *intext,
				 int size, unsigned char *key,
				 unsigned char *iv);
    do_decrypt = gcrypt_do_decrypt;
    crypto_ctx (*crypto_init) (struct options opt, int dir);
    crypto_init = gcrypt_crypto_init;
    void (*crypto_deinit) (crypto_ctx ctx);
    crypto_deinit = gcrypt_crypto_deinit;
#endif

    ctunnel = (struct Ctunnel *) arg;

    if (ctunnel->opt.proto == 0) {
	net_srv = calloc(1, sizeof(struct Network));
	net_cli = calloc(1, sizeof(struct Network));
	net_cli->sockfd = (int) ctunnel->clisockfd;
	net_srv->sockfd = (int) ctunnel->srvsockfd;
    } else {
	net_srv = (struct Network *) ctunnel->net_srv;
	net_cli = (struct Network *) ctunnel->net_cli;
    }

    ectx = crypto_init(ctunnel->opt, 1);
    dctx = crypto_init(ctunnel->opt, 0);


    if (ctunnel->opt.comp == 1)
	comp = z_compress_init(ctunnel->opt);

    data = calloc(1, sizeof(char *) * ctunnel->opt.packet_size);

    while (1) {
	if (ctunnel->opt.proto == 1) {
	    crypto_deinit(ectx);
	    crypto_deinit(dctx);
	    ectx = crypto_init(ctunnel->opt, 1);
	    dctx = crypto_init(ctunnel->opt, 0);
	}

	tv.tv_sec = 5;
	tv.tv_usec = 0;

	FD_ZERO(&rfd);
	FD_SET(net_srv->sockfd, &rfd);
	FD_SET(net_cli->sockfd, &rfd);
	ret = select(net_srv->sockfd + net_cli->sockfd + 1, &rfd,
		     NULL, NULL, &tv);
	if (ret < 0) 
	    break;

	if (FD_ISSET(net_srv->sockfd, &rfd)) {
	    ret = net_read(ctunnel->opt, net_srv,
		           (char *) data, ctunnel->opt.packet_size);
	    if (ret <= 0)
		break;

	    if (ctunnel->opt.server == 0) {
		if (ctunnel->opt.comp == 1) {
		    pkt_in = do_encrypt(ectx, data, ret,
			                ctunnel->opt.key.key,
					ctunnel->opt.key.iv);
		    pkt_out = z_compress(comp.zip, pkt_in.data,
					 pkt_in.len);
		} else {
		    pkt_out = do_encrypt(ectx, data, ret,
			                 ctunnel->opt.key.key,
					 ctunnel->opt.key.iv);
		}
	    } else {
		if (ctunnel->opt.comp == 1) {
		    pkt_in = z_uncompress(comp.unzip, data, ret);
		    pkt_out = do_decrypt(dctx, pkt_in.data,
					 pkt_in.len, ctunnel->opt.key.key,
					 ctunnel->opt.key.iv);
		} else {
		    pkt_out = do_decrypt(dctx, data, ret,
			                 ctunnel->opt.key.key,
					 ctunnel->opt.key.iv);
		}
	    }
	    ret = net_write(ctunnel->opt, net_cli,
		            (char *) pkt_out.data, pkt_out.len);

	    if (ctunnel->opt.comp == 1)
	        free(pkt_in.data);
	    free(pkt_out.data);

	    memset(data, 0x00, sizeof(char *) * ctunnel->opt.packet_size);

	    if (ret <= 0) {
		fprintf(stdout, "net_write %s:%d\n", __FILE__, __LINE__);
		break;
	    }
	}
	if (FD_ISSET(net_cli->sockfd, &rfd)) {
	    ret = net_read(ctunnel->opt, net_cli,
		           (char *) data, ctunnel->opt.packet_size);
	    if (ret <= 0) 
		break;

	    if (ctunnel->opt.server == 0) {
		if (ctunnel->opt.comp == 1) {
		    pkt_in = z_uncompress(comp.unzip, data, ret);
		    pkt_out = do_decrypt(dctx, pkt_in.data, pkt_in.len,
					 ctunnel->opt.key.key,
					 ctunnel->opt.key.iv);
		} else {
		    pkt_out = do_decrypt(dctx, data, ret,
					 ctunnel->opt.key.key,
					 ctunnel->opt.key.iv);
		}
	    } else {
		if (ctunnel->opt.comp == 1) {
		    pkt_in = do_encrypt(ectx, data, ret,
					ctunnel->opt.key.key,
					ctunnel->opt.key.iv);
		    pkt_out = z_compress(comp.zip, pkt_in.data, pkt_in.len);
		} else {
		    pkt_out = do_encrypt(ectx, data, ret,
					 ctunnel->opt.key.key,
					 ctunnel->opt.key.iv);
		}
	    }
	    ret = net_write(ctunnel->opt, net_srv,
		            (char *) pkt_out.data, pkt_out.len);

	    if (ctunnel->opt.comp == 1)
	        free(pkt_in.data);
	    free(pkt_out.data);

	    memset(data, 0x00, sizeof(char *) * ctunnel->opt.packet_size);

	    if (ret <= 0) {
		fprintf(stdout, "net_write %s:%d\n", __FILE__, __LINE__);
		break;
	    }
	}
	if (ctunnel->opt.proto == 1 && ctunnel->opt.comp == 1) {
	    deflateEnd(&comp.zip);
	    inflateEnd(&comp.unzip);
	    comp = z_compress_init(ctunnel->opt);
	}
	memset(data, 0x00, sizeof(char *) * ctunnel->opt.packet_size);
    }
    /* Connection closed */
    FD_CLR(net_cli->sockfd, &rfd);
    FD_CLR(net_srv->sockfd, &rfd);

    if (ctunnel->opt.proto == 0) {
	pthread_mutex_lock(&mutex);
	threads[ctunnel->id] = 2;
	pthread_mutex_unlock(&mutex);
	//fprintf(stdout, "THREAD: Exit %d\n", ctunnel->id);
	net_close(net_cli);
	net_close(net_srv);
	crypto_deinit(ectx);
	crypto_deinit(dctx);
	if (ctunnel->opt.comp == 1) {
	    deflateEnd(&comp.zip);
	    inflateEnd(&comp.unzip);
	}

	free(net_cli);
	free(net_srv);
    }
    free(data);
    free(ctunnel);
    pthread_exit(0);
    return NULL;
}
